import sys

import clr
clr.AddReferenceByPartialName("System.Xml")

from System import DateTime, Environment, Random
from System.IO import File, Path
from System.Xml import XmlDocument

from verb import Verb
from noun import Noun
from adjective import Adjective
from result import Result
from weightedselector import WeightedSelector

def random_element(lst):
    r = Random()
    return lst[int(r.NextDouble() * (len(lst)) - 1)]

class QuestionManager(object):
    def __init__(self, questions_path, results_path):
        self.questions_path = questions_path
        self.results_path = results_path
        self.questions = self._load_questions()
        self._attach_results(self.questions)
        
        self.answered_questions = []
        self.unanswered_questions = []
        
        for question in self.questions.values():
            if len(question.results) == 0:
                self.unanswered_questions.append(question)
            else:
                self.answered_questions.append(question)
        
    def _load_questions(self):
        questions = {}
        doc = XmlDocument()
        doc.Load(self.questions_path)
        
        for element in doc.GetElementsByTagName("noun"):
            n = Noun.from_element(element)
            
            assert not questions.has_key(n.id), str(n)
            questions[n.id] = n

        for element in doc.GetElementsByTagName("verb"):
            v = Verb.from_element(element)
            
            assert not questions.has_key(v.id), str(v)        
            questions[v.id] = v

        for element in doc.GetElementsByTagName("adjective"):
            a = Adjective.from_element(element)
            
            assert not questions.has_key(a.id), str(a)
            questions[a.id] = a
            
        return questions
            
    def _attach_results(self, questions):        
        if not File.Exists(self.results_path):
            return
        
        doc = XmlDocument()
        doc.Load(self.results_path)

        for element in doc.GetElementsByTagName("question"):
            id, lst = Result.result_list_from_element(element)
            
            if not questions.has_key(id):
                print >> sys.stderr, "missing key %r for results, discarding" % id
            else:
                questions[id].results = lst
                
                                    
    def save_results(self):
        document = XmlDocument()
        doc_elem = document.CreateElement("results")
        document.AppendChild(doc_elem)
        
        for question in self.answered_questions:
            doc_elem.AppendChild(question.create_result_association(document))
                    
        document.Save(self.results_path)

    def _should_introduce_new_question(self):
        poorly_answered_count = 0
        
        for question in self.answered_questions:
            if question.correctness() < 0.75:
                poorly_answered_count = poorly_answered_count + 1
                if poorly_answered_count > 10:
                    return False
        return True
    
    def _least_current_question_time(self):
        t = DateTime.Now
        for q in self.answered_questions:
            if len(q.results) > 0:
                t = min(t, q.results[-1].time)
                
        return t
    
    def currentness(self, question, min_time):
        if len(question.results) == 0:
            return 0
        else:
            return (question.results[-1].time - min_time).TotalSeconds / \
                   (DateTime.Now - min_time).TotalSeconds
    
    def weight(self, question, min_time):
        return (1 - question.correctness()) * \
               (1 - self.currentness(question, min_time))
    
    def get_question(self):
        if self._should_introduce_new_question():
            if len(self.unanswered_questions) == 0:
                print '*** add more questions ***'
            else:
                new_question = random_element(self.unanswered_questions)
                self.answered_questions.append(new_question)
                self.unanswered_questions.remove(new_question)
                return new_question

        ws = WeightedSelector()
        
        min_t = self._least_current_question_time()
        for q in self.answered_questions:
            ws.add(q, self.weight(q, min_t))
            
        print ws.print_weights()
        
        return ws.choose()
    
    def add_question(self, question):
        if self.questions.has_key(question.id):
            raise Exception("XXX")
        else:
            self.questions[question.id] = question
        
    def save_questions(self):
        questions = self.questions.values()
        questions.sort(lambda a, b: cmp(str(type(a)), str(type(b))))
        
        document = XmlDocument()
        doc_elem = document.CreateElement("xxx")
        document.AppendChild(doc_elem)
        
        for question in questions:
            doc_elem.AppendChild(question.to_element(document))
                                
        document.Save(self.questions_path)
        
def main():
    q = QuestionManager()
    
    try:
        while 1:
                q.get_question().ask_console()
                print
                print '*' * 79
                print
    finally:
        q.save()

if __name__ == '__main__':
    main()